@@ -1437,7 +1437,7 @@ func TestChunkAtBlockBoundary(t *testing.T) {
 	testutil.Ok(t, err)
 
 	for _, block := range db.Blocks() {
-		r, err := block.Index(math.MinInt64, math.MaxInt64)
+		r, err := block.Index()
 		testutil.Ok(t, err)
 		defer r.Close()
 
@@ -1787,7 +1787,7 @@ func TestDB_LabelNames(t *testing.T) {
 		appendSamples(db, 0, 4, tst.sampleLabels1)
 
 		// Testing head.
-		headIndexr, err := db.head.Index(math.MinInt64, math.MaxInt64)
+		headIndexr, err := db.head.Index()
 		testutil.Ok(t, err)
 		labelNames, err := headIndexr.LabelNames()
 		testutil.Ok(t, err)
@@ -1800,7 +1800,7 @@ func TestDB_LabelNames(t *testing.T) {
 		// All blocks have same label names, hence check them individually.
 		// No need to aggregate and check.
 		for _, b := range db.Blocks() {
-			blockIndexr, err := b.Index(math.MinInt64, math.MaxInt64)
+			blockIndexr, err := b.Index()
 			testutil.Ok(t, err)
 			labelNames, err = blockIndexr.LabelNames()
 			testutil.Ok(t, err)
