@@ -1,8 +1,8 @@
 import React, { FC } from 'react';
 import { RouteComponentProps } from '@reach/router';
-import PathPrefixProps from '../../PathPrefixProps';
-import { useFetch } from '../../utils/useFetch';
-import { withStatusIndicator } from '../../withStatusIndicator';
+import PathPrefixProps from '../../types/PathPrefixProps';
+import { useFetch } from '../../hooks/useFetch';
+import { withStatusIndicator } from '../../components/withStatusIndicator';
 import AlertsContent, { RuleStatus, AlertsProps } from './AlertContents';
 
 const AlertsWithStatusIndicator = withStatusIndicator(AlertsContent);
@@ -20,14 +20,7 @@ const Alerts: FC<RouteComponentProps & PathPrefixProps> = ({ pathPrefix = '' })
     response.data.groups.forEach(el => el.rules.forEach(r => ruleStatsCount[r.state]++));
   }
 
-  return (
-    <AlertsWithStatusIndicator
-      statsCount={ruleStatsCount}
-      groups={response.data && response.data.groups}
-      error={error}
-      isLoading={isLoading}
-    />
-  );
+  return <AlertsWithStatusIndicator statsCount={ruleStatsCount} {...response.data} error={error} isLoading={isLoading} />;
 };
 
 export default Alerts;
